//---------------------------------------------------------------------------
#include "tnlib.h"
#pragma hdrstop
#include <wtypes.h>		// for BSTR

#include "defs.h"
#include "Generic.h"
#include "WebDicDll.h"
#include "HTMLParser.h"
#include "sowUtil.h"
#include "pdstrlib.h"
#include "filestr.h"

#pragma package(smart_init)

#define	VOFILE_DIR		_t("VoiceFiles")
#define	VOFILE_PREFIX	_t("c2_")

inline bool strequ(const tchar *s1, const tchar *s2)
	{ return _tcscmp(s1, s2)==0; }
inline bool strequi(const tchar *s1, const tchar *s2)
	{ return _tcsicmp(s1, s2)==0; }
inline bool strequi(const tchar *s1, const tchar *s2, int n2)
	{ return _tcsnicmp(s1, s2, n2)==0; }

enum WDIndexes {
	INX_WEBLIO_EJ = 0,
	INX_WEBLIO_JP,
	INX_DICOVOX,
	INX_GOOGLE,
	INX_LOCAL,
};

TWDEnumerate Properties[] = {
	{ _t("Weblio(EiwaWaei)"), _t("!WeblioEiwaWaei"), "http://ejje.weblio.jp/content/", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ _t("Weblio(Japanese)"), _t("!WeblioJapanese"), "http://www.weblio.jp/content/", -1, WDF_IMMEDIATE|WDF_RELATED },
#ifdef _DEBUG
	{ _t("DicoVox(EF)"), _t("!DicoVox(EF)"), "http://129.194.19.89:8080/dicovox/index.php?dictionnaire=ef&rechercher=Rechercher / Find&mot=", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ _t("Google"), _t("!Google"), "http://www.google.co.jp/", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ _t("Local"), _t("!Local"), "http://127.0.0.1/", -1, WDF_IMMEDIATE|WDF_RELATED },
#endif
	{ NULL }
};

bool AudioDownload = true;
bool AudioMP3 = false;

//
// TGenericParser class
//
class TGenericParser : public THtmlStreamParser {
typedef THtmlStreamParser super;
protected:
	enum eSection {
		SEC_NONE,
		SEC_LANG,
		SEC_PRON,
		SEC_PART,
		SEC_SYN,	// Synonyms
		SEC_ANT,	// Antonyms
		SEC_CONTENTS=0x1000,
	};
	enum eTag {
		TAG_NONE,
		TAG_H1,
		TAG_H2,
		TAG_H3,
		TAG_H4,
		TAG_SPAN=0x10,
		TAG_OL=0x20,
		TAG_LI=0x40,
		TAG_DL=0x80,
		TAG_DD=0x100,
		TAG_UL=0x200,
		TAG_TABLE=0x1000,
	};
	enum eState {
		ST_NONE,
		ST_IPA,
		ST_AUDIOLINK,
		ST_HEADLINE,
		ST_TEXT,
		ST_DICNAME,
	};
protected:
	int Section;
	int Tag;
	int State;
	int LICounter;
	tnstr Language;	// current language
	tnstr Part;	// current part
	// parsed info. //
	struct ParsedInfo {
		tnstr ipa;
		tnstr audio_url;
		tnstr_map parts;
	};
	tnstr_vec Langs;		// languages
	tnstr_map LangNames;	// language -> language name (if exist)
	typedef map<tnstr, ParsedInfo> pimap_t;
	pimap_t Text;	// lang -> ParsedInfo mapping

	bool UseLF;

	// Formatter
public:
	static bool FmtPron;		// Embed phonetic symbols
	bool FmtAudioUrl;	// Embed Audio URL in the translation?
	bool FmtPartName;	// Embed the part name in the translation?
	
protected:
	virtual void startTag(const tchar *tagtop, const tchar *end);
	virtual void endTag(const tchar *tagtop, const tchar *end);
	virtual void startComment(const tchar *text)
	{
//		printf("start comment:%ls\n", text);
	}
	virtual void endComment(const tchar *end)
	{
//		printf("end comment:\n");
	}
	virtual void rawText(const tchar *text, int len);
	virtual void commentText(const tchar *text, int len)
	{
//		printf("comment:%ls\n", text);
	}
	void addText(tnstr &s, const tchar *text, int len);
	void addTextLF(tnstr &s);

	// new methods //
	virtual void h2Class(const tchar *name){}
	virtual void divClass(const tchar *name){}

	void setAudioUrl(const tchar *url);
public:
	TGenericParser()
	{
		Section = SEC_NONE;
		Tag = TAG_NONE;
		State = ST_NONE;
		UseLF = true;
		FmtAudioUrl = false;
		FmtPartName = false;
	}
	void dumpParsedInfo();
	void getParsedInfo(tnstrbuf &s, tnstrbuf &pron);
	tnstr getAudioUrl(int index);
};

bool TGenericParser::FmtPron = true;

void TGenericParser::startTag(const tchar *tagtop, const tchar *end)
{
	if (strequi(tagtop, _t("h1"), 2)){
		Tag = TAG_H1;
	} else
	if (strequi(tagtop, _t("h2"), 2)){
		Tag = TAG_H2;
		tnstr h2class = getTagAttr(tagtop, end, _t("class"));
		h2Class(h2class);
		//Language.clear();
		//Section = SEC_NONE;
	} else
	if (strequi(tagtop, _t("h3"), 2)){
		Tag = TAG_H3;
	} else
	if (strequi(tagtop, _t("h4"), 2)){
		Tag = TAG_H4;
	} else
	if (strequi(tagtop, _t("span"), 4)){
		Tag |= TAG_SPAN;
	} else
	if (State==ST_AUDIOLINK){
		if (Language.exist()){
			if (strequi(tagtop, _t("a href"), 6)){
				// audio url
				setAudioUrl(getTagAttr(tagtop, end, _t("href")));
			}
		}
	}
	// tags for contents
	if (Section & SEC_CONTENTS){
		if (strequi(tagtop, _t("ol"), 2)){
			Tag |= TAG_OL;
			LICounter = 0;
		} else
		if (strequi(tagtop, _t("ul"), 2)){
			Tag |= TAG_UL;
		} else
		if (strequi(tagtop, _t("li"), 2)){
			Tag |= TAG_LI;
			if (Tag & TAG_OL){
				tchar buf[20];
				buf[0] = '\n';
				itos(++LICounter, &buf[1]);
				_tcscat(buf, _t("."));
				addText(Text[Language].parts[Part], buf, _tcslen(buf));
			} else
			if (Tag & TAG_UL){
				tchar buf[6];
				_tcscpy(buf, _t("\n - "));
				addText(Text[Language].parts[Part], buf, _tcslen(buf));
			}
		} else
		if (strequi(tagtop, _t("dl"), 2)){
			Tag |= TAG_DL;
		} else
		if (strequi(tagtop, _t("dd"), 2)){
			Tag |= TAG_DD;
			addTextLF(Text[Language].parts[Part]);
		} else
		if (strequi(tagtop, _t("br"), 2)){
			addTextLF(Text[Language].parts[Part]);
		} else
		if (strequi(tagtop, _t("table"), 5)){
			Tag |= TAG_TABLE;
		} else
		if (strequi(tagtop, _t("div"), 3)){
			tnstr divclass = getTagAttr(tagtop, end, _t("class"));
			if (divclass.exist()){
				divClass(divclass);
			}
		}
	}
}

void TGenericParser::endTag(const tchar *tagtop, const tchar *end)
{
	if (Tag==TAG_NONE)
		return;

	if (strequi(tagtop, _t("h1"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h2"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h3"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h4"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("span"), 4)){
		Tag &= ~TAG_SPAN;
	}
	// tags for contents
	if (Section & SEC_CONTENTS){
		if (strequi(tagtop, _t("ol"), 2)){
			Tag &= ~TAG_OL;
		} else
		if (strequi(tagtop, _t("ul"), 2)){
			Tag &= ~TAG_UL;
		} else
		if (strequi(tagtop, _t("li"), 2)){
			Tag &= ~TAG_LI;
		} else
		if (strequi(tagtop, _t("dl"), 2)){
			Tag &= ~TAG_DL;
		} else
		if (strequi(tagtop, _t("dd"), 2)){
			Tag &= ~TAG_DD;
		} else
		if (strequi(tagtop, _t("table"), 5)){
			Tag &= ~TAG_TABLE;
		}
	}
}

void TGenericParser::rawText(const tchar *text, int len)
{
	if (Section==SEC_NONE && Tag==TAG_NONE)
		return;

	tnstr s;
	switch (Tag){
		case TAG_H2|TAG_SPAN:
			if (State==ST_HEADLINE){
				Language.set(text, len);
				Langs.add(Language);
				//printf("lang:%ls\n", Language.c_str());
				if (Language.exist()){
					Section = SEC_LANG;
				} else {
					Section = SEC_NONE;
				}
			}
			break;
		case TAG_H3|TAG_SPAN:
			if (State==ST_HEADLINE){
				//_tprintf(_t("head:%s\n"), tnstr(text,len).c_str());
				if (Language.exist()){
					Part.set(text, len);
					if (Part==_t("Pronunciation")){
						Section = SEC_PRON;
					} else {
						if (Part.exist()){
							Section = SEC_PART|SEC_CONTENTS;
						} else {
							Section = SEC_LANG;
						}
					}
				}
			}
			break;
		case TAG_H4|TAG_SPAN:
			if (State==ST_HEADLINE){
				_tprintf(_t("head:%s\n"), tnstr(text,len).c_str());
				if (Language.exist()){
					s.set(text, len);
					if (s==_t("Synonyms")){
						Section = SEC_SYN|SEC_CONTENTS;
						Part = s;
					} else
					if (s==_t("Antonyms")){
						Section = SEC_ANT|SEC_CONTENTS;
						Part = s;
					} else {
						Section = SEC_LANG;
					}
				}
			}
			break;
		default:
			if (State==ST_IPA){
				if (Language.exist()){
					if (Text[Language].ipa.empty()){
						Text[Language].ipa.set(text, len);	// IPA
						//printf("ipa:%ls\n", Text[Language].ipa.c_str());
					}
				}
			}
			if (Section&SEC_CONTENTS){
				// plain text
				addText(Text[Language].parts[Part], text, len);
			}
			break;
	}
}

void TGenericParser::addText(tnstr &s, const tchar *text, int len)
{
	if (Tag & TAG_TABLE)
		return;
	if (s.exist()){
		tchar last = s[s.length()-1];
		if (text[0]=='\n'){
			if (last=='\n'){
				text++;
				len--;
			}
		} else {
			if (last!=' '){
				//s += _t(" ");
			}
		}
	} else {
		while (text[0]=='\n'){
			text++;
			len--;
		}
	}
	s += UnescapeHtml(text, len);
	//s += tnstr(text, len);
}

void TGenericParser::addTextLF(tnstr &s)
{
	if (UseLF){
		addText(s, _t("\n"), 1);
	}
}

void TGenericParser::setAudioUrl(const tchar *url)
{
	if (Language.empty()) return;
	if (!url[0]) return;
	if (Text[Language].audio_url.empty()){
		Text[Language].audio_url = url;
	}
}

#ifdef _DEBUG
// for debug
void TGenericParser::dumpParsedInfo()
{
	for(pimap_t::iterator it=Text.begin();it!=Text.end();it++){
		_tprintf(_t("----%s\n"), it->first.c_str());
		ParsedInfo &pi = it->second;
		_tprintf(_t("ipa  :%s\n"), pi.ipa.c_str());
		_tprintf(_t("audio:%ls\n"), pi.audio_url.c_str());
		foreach(pi.parts, it, tnstr_map){
			_tprintf(_t("[%ls]\n"), it->first.c_str());
			_tprintf(_t("%ls\n"), it->second.c_str());
		}
	}
}
#endif	// _DEBUG

void TGenericParser::getParsedInfo(tnstrbuf &s, tnstrbuf &pron)
{
	foreach_tnstr_vec(Langs, it){
		ParsedInfo &pi = Text[*it];
		s += _t("<b>");
		if (LangNames.count(*it)){
			s += LangNames[*it];
		} else {
			s += *it;
		}
		s += _t("</b>\n");

		if (pi.ipa.exist()){
			s += _t("ipa  :");
			s += pi.ipa;
			//TODO: ŏĂHHs += _t("\n");
		}
		if (FmtAudioUrl){
			if (pi.audio_url.exist()){
				s += _t("audio:");
				s += pi.audio_url;
				s += _t("\n");
			}
		}
		foreach(pi.parts, it, tnstr_map){
			if (it->first.exist()){
				if (strequ(it->first, _t("pron"))){
					if (FmtPron){
						if (pron.exist()){
							pron += _t(",");	// delimiter for phonetic symbol
						}
						pron += it->second;
					}
					continue;
				}
				if (FmtPartName){
					s += _t("[");
					s += it->first;
					s += _t("]\n");
				}
				s += it->second;
//				s += _t("\n");
			}
		}
	}
}

tnstr TGenericParser::getAudioUrl(int index)
{
	if (index<0 || index>=Langs.size())
		return NULL;
	if (Langs[index].empty() || Text[Langs[index]].audio_url.empty())
		return _t("\x01");	// no url
	return Text[Langs[index]].audio_url;
}

const tchar *TAG_NAME = _t("name");
static const tchar *oto = _t("");

//
// TWeblioParser class
//
class TWeblioParser : public TGenericParser {
typedef TGenericParser super;
protected:
	tnstr cur_dicTag;		// <a name="xxx">
	tnstr_vec PartStack;	// <div>...</div> stack
	bool TextOut;
public:
	bool AudioMP3;	// download audio for mp3.
public:
	TWeblioParser()
	{
		TextOut = false;
		AudioMP3 = ::AudioMP3;
	}
protected:
	virtual void startTag(const tchar *tagtop, const tchar *end)
	{
		if (strequi(tagtop, _t("param"), 5)){
			tnstr_map attrs;
			parseTag(tagtop, end, attrs);
			if (attrs.count(TAG_NAME) && strequ(attrs[TAG_NAME], _t("src")) && attrs.count(_t("value"))){
				tnstr url = attrs[_t("value")];
				if (_tcsstr(url, _t(".wav"))){
					// audio url (obsolete)
					// <param name="src" value="xxx.wav">
					setAudioUrl(url);
				}
			}
			
		} else
		if (strequi(tagtop, _t("a "), 2)){
			tnstr_map attrs;
			parseTag(tagtop, end, attrs);
			if (attrs.count(TAG_NAME)){
				// <a name="xxx">
				cur_dicTag = attrs[TAG_NAME];
				Section &= ~SEC_CONTENTS;
			}
		} else
		if (strequi(tagtop, _t("span"), 4)){
			tnstr_map attrs;
			parseTag(tagtop, end, attrs);
			if (attrs.count(_t("onclick"))){
				tnstr onclick = attrs[_t("onclick")];
				const tchar *playsound = _tcsstr(onclick, _t("PlaySound('"));
				if (playsound){
					playsound += 11;
					const tchar *end = _tcschr(playsound, '\'');
					if (end){
						tnstr url(playsound, (end-playsound));
						if (AudioMP3){
							url = tnstr(_t("http://www.weblio.jp/img/dict/kenej/audio/"))+url+_t(".mp3");
						} else {
							url = tnstr(_t("http://www.weblio.jp/wbr/CHUJITEN/"))+url+_t(".wav");
						}
						setAudioUrl(url);
					}
				}
			}
			//
		}
		super::startTag(tagtop, end);
	}
	virtual void h2Class(const tchar *name)
	{
		if (strequ(name, _t("midashigo"))){
			Section |= SEC_CONTENTS;
//			Langs.add(_t("def"));
//			Language = _t("def");
//			Part = _t("text");
		} else
		if (strequ(name, _t("dictNm"))){
			State |= ST_DICNAME;
		}
	}
	virtual void divClass(const tchar *name)
	{
		PartStack.push_back(name);
		enterDiv(name);

		// Additional text to be easy to see.
		if (Part.exist()){
			if (strequ(name, _t("reibun_folding"))
				|| strequ(name, _t("reibun_first"))
				|| strequ(name, _t("KejjeYrLn"))
				){
				addText(Text[Language].parts[Part], _t("- "), 3);
			}
		}
	}
	virtual void enterDiv(const tchar *name)
	{
		if (TextOut){
			//assert(Part.exist());
			addTextLF(Text[Language].parts[Part]);
		}
		if (IsTextClass(name)){
			State |= ST_TEXT;
			if (strequ(name, _t("Kejje"))){
				Part = _t("pron");	// phonetic symbol
				UseLF = false;
			} else {
				Part = _t("text");
				UseLF = true;
			}
		} else {
			State &= ~ST_TEXT;
			Part.clear();
		}
		TextOut = false;
	}
	static bool IsTextClass(const tchar *name)
	{
		if (strequ(name, _t("level0"))
			|| strequ(name, _t("level1"))
			|| strequ(name, _t("level2"))
			|| strequ(name, _t("Kejje"))	// Phonetic symbol
			//|| strequ(name, _t("KejjeYrTtlFld")) "p"
			|| strequ(name, _t("KejjeYrHd"))
			|| strequ(name, _t("KejjeYrLn"))
			|| strequ(name, _t("KejjeYrTxt"))
			|| strequ(name, _t("KejjeYrTtl"))
			
			|| strequ(name, _t("Liscj"))
			|| strequ(name, _t("LiscjYr"))
			//|| strequ(name, _t("LiscjYrHd"))	"֘A"
			|| strequ(name, _t("kanren"))
			|| strequ(name, _t("LiscjYr"))
			|| strequ(name, _t("LiscjYrHdFld"))
			|| strequ(name, _t("reibun_first"))
			|| strequ(name, _t("reibun_folding"))

			|| strequ(name, _t("Wejty"))

			// pawpW
			|| strequ(name, _t("Eigky"))

			// Oȓ 厫
			|| strequ(name, _t("NetDicBody"))
			){
			return true;
		}
		return false;
	}
	virtual void rawText(const tchar *text, int len)
	{
		if (Section&SEC_CONTENTS && State&ST_TEXT){
			// plain text
			if (Language.exist() && Part.exist()){
				if (len>0){
					if (Part==_t("pron")){
						// "" "L" "𕷂"菜
						// ȗ""ĂΑΏۊO
						// C++Builder̃oÔstaticϐœn
						if (_tcsnstr(text, oto, len)){
							return;
						}
					}
					addText(Text[Language].parts[Part], text, len);
					TextOut = true;
				}
			}
		}
		if (State & ST_DICNAME){
			if (cur_dicTag.exist()){
				// <h2 class=dictNm>xxx</h2>
				tnstr title(text, len);
				LangNames[cur_dicTag] = title;
				Langs.add(cur_dicTag);
				Language = cur_dicTag;
				cur_dicTag.clear();
			}
		}
	}
	virtual void endTag(const tchar *tagtop, const tchar *end)
	{
		if (strequi(tagtop, _t("div"), 3)){
			// exit div
			State &= ~(ST_TEXT|ST_DICNAME);

			// pop the Part
			if (PartStack.size()>0){
				PartStack.pop();
				if (PartStack.size()>0){
					enterDiv(PartStack[PartStack.get_num()-1]);
				} else {
					enterDiv(_t(""));
				}
			}
		}
		super::endTag(tagtop, end);
	}
};

//
// TDicoVoxParser class
//
class TDicoVoxParser : public TGenericParser {
typedef TGenericParser super;
};


int genRequestData(int index, const tchar *srch_word, TWDResponse *resp, const tchar *dictionary_name, TWDInetAccess *inet)
{
	tnstrA srchU;
	{
		tnstrA temp;
		temp.setUTF8(srch_word);	// UTF-16->UTF-8
		EscapeURIString(temp, srchU);	// escape
	}

	__tnstrA url;
	url += Properties[index].url;
	url += srchU;
	BSTR res;
	if (!inet->HttpReq(url, &res)){
		return 0;
	}
	tnstr text1;
	tnstrbuf text2;
	text1.setUTF8((char*)res);
	SysFreeString(res);

	const tchar *text = text1;
			
	if (*resp->cancelReq){
		return 0;
	}

	bool ok = false;
#if 0
	TOFile file;
	file.create(_t("capture.html"));
	file.putline(text1);
	file.close();
#endif
	TGenericParser *parser;
	switch (index){
		case INX_WEBLIO_EJ:
		case INX_WEBLIO_JP:
		case INX_LOCAL:
			parser = new TWeblioParser();
			break;
		case INX_DICOVOX:
			parser = new TDicoVoxParser();
			break;
		default:
			parser = new TGenericParser();
			break;
	}

	tnstrbuf text_pron;

	//THtmlStreamParser parser;
	if (parser->openBuffer(text1)){
		parser->parse();
		parser->getParsedInfo(text2, text_pron);
		if (text2.exist()){
			text = text2;
			ok = true;
#ifdef MODULE_TEST
			{
				TOFile file;
				file.create(_t("genout.txt"));
				if (text_pron.exist()){
					file.put(_t("PRON:"));
					file.putline(text_pron);
				}
				file.put(text2);
				file.close();
			}
#endif

			if (AudioDownload){
				// Audio download
				int findex = 0;
				for (int i=0;;i++){
					tnstr audio_url = parser->getAudioUrl(i);
					if (audio_url.empty())
						break;
					if (audio_url[0]==0x01)
						continue;
					// t@C͏㏑
					__tnstrA audio_urlA; audio_urlA.setUTF8(audio_url);
					__tnstrA voiceurl = MakeUrl(url, audio_urlA);
					tnstr voicefile = MakeVoiceFile(dictionary_name, audio_urlA, VOFILE_DIR, VOFILE_PREFIX);
					tnstr linkname = GetFileName(voicefile);

					if (inet->GetHttpFile(voiceurl, "", voicefile)==0
						|| fexist(voicefile)	//Note: voicefileقŎgpŃG[ɂȂ\邽
						){
						resp->filelinks[findex++] = SysAllocString(linkname);
					}
				}
			}
		}
	}
	delete parser;

	if (*resp->cancelReq || !ok){
		return 0;
	}

	resp->japa = SysAllocString(text);
	if (text_pron.exist()){
		resp->pron = SysAllocString(text_pron);
	}

	return 1;
}

#define	profInt(keyname, value)	GetPrivateProfileInt(_t("config"), keyname, value, IniName)

void genInit(HINSTANCE hinst)
{
	const tchar *IniName = _t("GenericWD.ini");
	TCHAR module[300];
	if (GetModuleFileName(hinst, module, sizeof(module)/sizeof(TCHAR))>0){
		GetFileExtension(module)[0] = '\0';
		_tcscat(module, _t("ini"));
		IniName = module;
	}

	AudioDownload = profInt(_t("AudioEnabled"), AudioDownload);
	AudioMP3 = profInt(_t("AudioMP3"), AudioMP3);
	TGenericParser::FmtPron = profInt(_t("PronEnabled"), TGenericParser::FmtPron);
}

