#include "tnlib.h"
#pragma hdrstop

#include "pdstrlib.h"
#include "HtmlParser.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

bool THtmlParser::parse()
{
	int state = 0;	// 1:in tag 2:in tag/quote 4:in comment
	int type = 0;
	const tchar *tagtop = NULL;

	tnstr tag;
	tnstr line;
	while (!AbortReq && getline(line)){
		const tchar *sp = line;
		const tchar *text = (state==0?sp:NULL);
		const tchar *comment = (state==4 ? sp : NULL);
		while (*sp){
			if (state==0){
				if (*sp=='<'){
					if (text){
						//text = skipspc(text);	// skip blank text
						if (text!=sp){
							rawText(text, STR_DIFF(sp,text));
						}
						text = NULL;
					}
					sp++;
					// start of HTML tag
					if (sp[0]=='!' && sp[1]=='-' && sp[2]=='-'){
						// start comment
						state = 4;
						sp += 3;
						comment = sp;
						startComment(sp);
						continue;
					} else {
						if (*sp=='/'){
							type = 1;
							sp++;
						} else {
							type = 0;
						}
						state = 1;
						sp = skipspc(sp);
						tagtop = sp;
					}
					continue;
				}
				if (!text){
					text = sp;
				}
			} else
			if (state>=4){
				// comment
				if (sp[0]=='-' && sp[1]=='-' && sp[2]=='>'){
					// end of comment
					if (comment){
						comment = skipspc(comment);
						if (comment!=sp){
							commentText(comment, STR_DIFF(sp,comment));
						}
						comment = NULL;
					}
					endComment(sp);
					state = 0;
					sp += 3;
					continue;
				}
			} else
			if (state==1){
				if (*sp=='>'){
					// end of HTML tag
					if (tagtop){
						if (type==0){
							startTag(tagtop, sp);
						} else {
							endTag(tagtop, sp);
						}
					}
					state = 0;
				} else
				if (*sp=='"'){
					state = 2;
				}
			} else
			if (state==2){
				// in tag/quote
				if (*sp=='"'){
					state = 1;
				}
			}
			sp++;
		}
		if (state==1 || state==2){
			//TODO: not yet
			fprintf(stderr,"WARNING!! - open tag\n");
			if (tagtop){
				if (type==0){
					startTag(tagtop, sp);
				} else {
					endTag(tagtop, sp);
				}
				tagtop = NULL;
			}
		} else
		if (text){
			if (sp!=text)
				rawText(text, STR_DIFF(sp, text));
		} else
		if (comment){
			if (sp!=comment)
				commentText(comment, STR_DIFF(sp, comment));
		}
	}
	return true;
}

void THtmlParser::parseTag(const tchar *tag, const tchar *end, tnstr_map &attrs)
{
	// skip tag name
	const tchar *sp = skiptospc(tag);
	sp = skipspc(sp);
	int state = 0;
	const tchar *name_top;
	const tchar *name_end = NULL;
	const tchar *value_top = NULL;
	tchar quote_char = 0;
	for (;sp<end;){
		if (state==0){
			if (isword(*sp)){
				// start of attribute name
				state = 1;
				name_top = sp;
			}
		} else
		if (state==1){
			if (!isword(*sp)){
				// end of attr name
				name_end = sp;
				sp = skipspc(sp);
				if (*sp=='='){
					state = 2;
				} else {
					// no value
					state = 0;
					attrs[tnstr(name_top, STR_DIFF(sp, name_top))] = NULL;
					name_end = NULL;
					continue;
				}
			}
		} else
		if (state==2){
			if (*sp=='"' || *sp=='\''){
				// quote
				state = 3;
				quote_char = *sp;
				value_top = sp+1;
			} else
			if (*sp!=' ' && *sp!='\t'){
				state = 4;
				value_top = sp;
			}
		} else
		if (state==3){
			// quoted text
			if (*sp==quote_char){
				state = 0;
				attrs[tnstr(name_top, STR_DIFF(name_end, name_top))] = tnstr(value_top, STR_DIFF(sp, value_top));
				name_end = NULL;
				value_top = NULL;
			}
		} else
		if (state==4){
			// bare text
			if (*sp==' ' || *sp=='\t'){
				state = 0;
				attrs[tnstr(name_top, STR_DIFF(name_end, name_top))] = tnstr(value_top, STR_DIFF(sp, value_top));
				name_end = NULL;
				value_top = NULL;
			}
		}
		sp++;
	}
	if (value_top){
		attrs[tnstr(name_top, STR_DIFF(name_end, name_top))] = tnstr(value_top, STR_DIFF(sp, value_top));
	} else
	if (name_end){
		attrs[tnstr(name_top, STR_DIFF(name_end, sp))] = NULL;	// no value
	}
}

tnstr THtmlParser::getTagAttr(const tchar *tag, const tchar *tag_end, const tchar *attr_name)
{
	tnstr_map attrs;
	parseTag(tag, tag_end, attrs);
	if (attrs.count(attr_name)){
		return attrs[attr_name];
	} else {
		return NULL;
	}
}

// THtmlStreamParser class //
bool THtmlStreamParser::open(const tchar *filename)
{
	if (File.open(filename)){
		return false;
	}
	return true;
}
bool THtmlStreamParser::openBuffer(const tchar *str)
{
	Buffer = str;
	return true;
}
bool THtmlStreamParser::getline(tnstr &s)
{
	if (Buffer){
		const tchar *p = Buffer;
		const tchar *end = NULL;
		while (*p){
			if (*p=='\r'){
				end = p;
				p++;
				if (*p=='\n'){
					p++;
				}
				break;
			} else
			if (*p=='\n'){
				end = p;
				p++;
				break;
			}
			p++;
		}
		if (p==Buffer)
			return false;	// end of buffer
		if (!end) end = p;
		s.set(Buffer, STR_DIFF(end,Buffer));
		Buffer = p;
		return true;
	}
	return File.getline(s)>=0;
}
#if 0	// for debug
void THtmlStreamParser::startTag(const tchar *tagtop, const tchar *end)
{
	tnstr s(tagtop, (int)(end-tagtop));
	_tprintf(_t("<%s>\n"), s.c_str());
}
void THtmlStreamParser::endTag(const tchar *tagtop, const tchar *end)
{
	tnstr s(tagtop, (int)(end-tagtop));
	_tprintf(_t("</%s>\n"), s.c_str());
}
void THtmlStreamParser::startComment(const tchar *text)
{
	_tprintf(_t("<!--\n"));
}
void THtmlStreamParser::endComment(const tchar *end)
{
	_tprintf(_t("-->\n"));
}
void THtmlStreamParser::rawText(const tchar *text, int len)
{
	tnstr s(text, len);
	_tprintf(_t("%s\n"), s.c_str());
}
void THtmlStreamParser::commentText(const tchar *text, int len)
{
	tnstr s(text, len);
	_tprintf(_t("%s\n"), s.c_str());
}
#endif


