//---------------------------------------------------------------------------
// Sub set of '../../WebDicInfo.cpp'
#include "tnlib.h"
#pragma hdrstop
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "WebDicInfoS.h"
//---------------------------------------------------------------------------

typedef bool (*FNHttpFile)(struct THttpFileParam &param);
typedef bool (*FNHttpIO)( char *data, int *datalen, int userdata );

struct THttpFileParam {
	int FileHandle;
	const tchar *FileName;
	int TotalSize;
	int LoadedSize;
	FNHttpFile Callback;
	int UserData;
};

// Prototype //
static bool HttpReq(const char *_url, __tnstrA &res);
static int HttpReq(const char *_url, BSTR *res);
static int GetHttpFile(const char *url, const char *param, const tchar *path);
static int InetDicDll_GetHttp(const char *url, const char *param, __tnstrA &res);
static int InetDicDll_GetHttpFile(const char *url, const char *param, const tchar *path, FNHttpFile cb=0, int userdata=0);
static int _RequestHttp(const char *url, const char *param, __tnstrA &res, bool post, FNHttpIO callback=0, int userdata=0);
static int InetDicDll_RequestHttp( const char *url, const char *proxy, int proxyPort, const char *param, bool post, int *status, FNHttpIO callback, int userdata, const char *useragent );
static int InetDicDll_GetHttpRes( int h_res, char *buf, int bufsize );
static void InetDicDll_DeleteHttpRes( int h_res );

#define	DEF_INETDIC( ret, name, args ) \
	typedef ret (*FN##name) args

DEF_INETDIC( int, RequestHttp, ( const char *url, const char *proxy, int proxyPort, const char *param, bool post, int *status, FNHttpIO callback, int userdata, const char *useragent ) );
DEF_INETDIC( int, GetHttpRes, ( int h_res, char *buf, int bufsize ));
DEF_INETDIC( void, DeleteHttpRes, ( int h_res ) );

const char *UserAgent = "PDIC/5 of Mozilla/4 compatible";

// -----------------------------------------------------------------------
//
// TWebDicExtInfo
//
TWDResponse::TWDResponse()
{
	memset(this, 0, sizeof(*this));
}

TWDResponse::~TWDResponse()
{
	if (pron) SysFreeString(pron);
	if (japa) SysFreeString(japa);
	if (exp) SysFreeString(exp);
	for (int i=0;i<16;i++){
		if (filelinks[i])
			SysFreeString(filelinks[i]);
	}
}
TWDInetAccess::TWDInetAccess()
{
	HttpReq = ::HttpReq;
	GetHttpFile = ::GetHttpFile;
}

static bool HttpReq(const char *_url, __tnstrA &res)
{
	const char *url = _url;
	const char *obj = "";
	const char *p = strchr(url, '?');
	__tnstrA urlBuf;
	if (p){
		obj = p+1;
		urlBuf.set(url, FP_DIFF(p, url));
		url = urlBuf;
	}
	if (InetDicDll_GetHttp(url, obj, res)){
		return false;
	}
	return true;
}

static int HttpReq(const char *_url, BSTR *res)
{
	*res = NULL;

	__tnstrA str;
	bool ret = HttpReq(_url, str);
	if (!ret)
		return 0;
	*res = SysAllocStringByteLen(str, str.length());
	return 1;
}

static int GetHttpFile(const char *url, const char *param, const tchar *path)
{
	return InetDicDll_GetHttpFile(url, param, path);
}

//
// Altenative methods for InetDicDll
//
int InetDicDll_GetHttp(const char *url, const char *param, __tnstrA &res)
{
	return _RequestHttp(url, param, res, false);
}

bool cbFileSaveStream( char *data, int *datalen, int userdata )
{
	THttpFileParam &hfp = *(THttpFileParam*)userdata;
	if (!data){
		// total size
		if (datalen)
			hfp.TotalSize = *datalen;
		return true;
	}
	if (write(hfp.FileHandle, data, *datalen)==-1){
		// error
		return false;
	}
	if (hfp.Callback){
		hfp.LoadedSize += *datalen;
		return hfp.Callback(hfp);
	}
	return true;
}

//	-3 : file create error
int InetDicDll_GetHttpFile(const char *url, const char *param, const tchar *path, FNHttpFile cb, int userdata)
{
	THttpFileParam hfp;
	__tnstrA pathA;
	pathA.setAnsi(path);
	int fd = creat(pathA, S_IREAD|S_IWRITE);
	if (fd<0){
		return -3;
	}
	hfp.FileHandle = fd;
	hfp.FileName = path;
	hfp.TotalSize = 0;
	hfp.LoadedSize = 0;
	hfp.Callback = cb;
	hfp.UserData = userdata;
	__tnstrA res;
	return _RequestHttp(url, param, res, false, cbFileSaveStream, (int)&hfp);
}

int _RequestHttp(const char *url, const char *param, __tnstrA &res, bool post, FNHttpIO callback, int userdata)
{
	int proxy_port = 0;
	int status = 0;
	int ret = InetDicDll_RequestHttp(url, "", proxy_port, param, post, &status, callback, userdata, UserAgent);
	if (ret==0){
		// no memory
		return 0;	//TODO: return value tbd.
	} else
	if (ret==-1 || ret==-2){
		return status;
	}
	if (status){
		char *buf = new char[status+1];
		InetDicDll_GetHttpRes(ret, buf, status+1);
		res.setBuf(buf);
	} else {
		res.clear();
	}
	InetDicDll_DeleteHttpRes(ret);
	return 0;
}

static HINSTANCE hInetDicDll = NULL;
static const tchar *InetDicDllName = _t("../../../InetDic.dll");
static FNRequestHttp RequestHttp;
static FNGetHttpRes GetHttpRes;
static FNDeleteHttpRes DeleteHttpRes;
static bool LoadInetDicDll()
{
	if (hInetDicDll)
		return true;
	hInetDicDll = LoadLibrary(InetDicDllName);
	if (!hInetDicDll){
		printf("Cannot load %S.\n", InetDicDllName);
		return false;
	}
	RequestHttp = (FNRequestHttp)GetProcAddress(hInetDicDll, "RequestHttp");
	GetHttpRes = (FNGetHttpRes)GetProcAddress(hInetDicDll, "GetHttpRes");
	DeleteHttpRes = (FNDeleteHttpRes)GetProcAddress(hInetDicDll, "DeleteHttpRes");
	return true;
}
void UnloadInetDicDll()
{
	if (hInetDicDll){
		FreeLibrary(hInetDicDll);
		hInetDicDll = NULL;
	}
}

static int InetDicDll_RequestHttp( const char *url, const char *proxy, int proxyPort, const char *param, bool post, int *status, FNHttpIO callback, int userdata, const char *useragent )
{
	if (!LoadInetDicDll())
		return -1;
	return RequestHttp(url, proxy, proxyPort, param, post, status, callback, userdata, useragent);
}

static int InetDicDll_GetHttpRes( int h_res, char *buf, int bufsize )
{
	if (!LoadInetDicDll())
		return -1;
	return GetHttpRes(h_res, buf, bufsize);
}
static void InetDicDll_DeleteHttpRes( int h_res )
{
	if (!LoadInetDicDll())
		return;
	DeleteHttpRes(h_res);
}

