//---------------------------------------------------------------------------
#include "tnlib.h"
#pragma hdrstop
#include <oleauto.h>

#include "defs.h"
#include "Wiktionary.h"
#include "WebDicDll.h"
#include "HTMLParser.h"
#include "sowUtil.h"

#include "pdstrlib.h"
#include "filestr.h"

#pragma package(smart_init)

#define	VOFILE_DIR		_t("VoiceFiles")
#define	VOFILE_PREFIX	_t("c1_")

inline bool strequ(const tchar *s1, const tchar *s2)
	{ return _tcscmp(s1, s2)==0; }
inline bool strequi(const tchar *s1, const tchar *s2)
	{ return _tcsicmp(s1, s2)==0; }
inline bool strequi(const tchar *s1, const tchar *s2, int n2)
	{ return _tcsnicmp(s1, s2, n2)==0; }

TWDEnumerate Properties[] = {
	{ _t("Wiktionary(JA)"), _t("!WiktionaryJA"), "http://ja.wiktionary.org/wiki/", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ _t("Wiktionary(EN)"), _t("!WiktionaryEN"), "http://en.wiktionary.org/wiki/", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ _t("Wiktionary(FR)"), _t("!WiktionaryFR"), "http://fr.wiktionary.org/wiki/", -1, WDF_IMMEDIATE|WDF_RELATED },
	{ NULL }
};

bool AudioDownload = true;

//
// TWiktionaryParser class
//
class TWiktionaryParser : public THtmlStreamParser {
typedef THtmlStreamParser super;
protected:
	enum eSection {
		SEC_NONE,
		SEC_LANG,
		SEC_PRON,
		SEC_PART,
		SEC_SYN,	// Synonyms
		SEC_ANT,	// Antonyms
		SEC_CONTENTS=0x1000,
	};
	enum eTag {
		TAG_NONE,
		TAG_H1,
		TAG_H2,
		TAG_H3,
		TAG_H4,
		TAG_SPAN=0x10,
		TAG_OL=0x20,
		TAG_LI=0x40,
		TAG_DL=0x80,
		TAG_DD=0x100,
		TAG_UL=0x200,
		TAG_TABLE=0x1000,
	};
	enum eState {
		ST_NONE,
		ST_IPA,
		ST_AUDIOLINK,
		ST_HEADLINE,
	};
protected:
	int Section;
	int Tag;
	int State;
	bool EditSection;
	int LICounter;
	tnstr Language;	// current language
	tnstr Part;	// current part
	// parsed info. //
	struct ParsedInfo {
		tnstr ipa;
		tnstr audio_url;
		tnstr_map parts;
	};
	tnstr_vec Langs;	// languages
	typedef map<tnstr, ParsedInfo> pimap_t;
	pimap_t Text;	// lang -> ParsedInfo mapping
	// Formatter
public:
	static bool FmtPron;		// Embed phonetic symbols
	bool FmtAudioUrl;	// Embed Audio URL in the translation?
	bool FmtPartName;	// Embed the part name in the translation?
	
protected:
	virtual void startTag(const tchar *tagtop, const tchar *end);
	virtual void endTag(const tchar *tagtop, const tchar *end);
	virtual void startComment(const tchar *text)
	{
//		printf("start comment:%ls\n", text);
	}
	virtual void endComment(const tchar *end)
	{
//		printf("end comment:\n");
	}
	virtual void rawText(const tchar *text, int len);
	virtual void commentText(const tchar *text, int len)
	{
	}
	void addText(tnstr &s, const tchar *text, int len);
	void addTextLF(tnstr &s);
	void setAudioUrl(const tchar *url);
public:
public:
	TWiktionaryParser()
	{
		Section = SEC_NONE;
		Tag = TAG_NONE;
		State = ST_NONE;
		EditSection = false;
		FmtAudioUrl = true;
		FmtPartName = true;
	}
	void dumpParsedInfo();
	void getParsedInfo(tnstrbuf &s, tnstrbuf &pron);
	tnstr getAudioUrl(int index);
};

bool TWiktionaryParser::FmtPron = false;

void TWiktionaryParser::startTag(const tchar *tagtop, const tchar *end)
{
	if (strequi(tagtop, _t("h1"), 2)){
		Tag = TAG_H1;
	} else
	if (strequi(tagtop, _t("h2"), 2)){
		Tag = TAG_H2;
		Language.clear();
		Section = SEC_NONE;
	} else
	if (strequi(tagtop, _t("h3"), 2)){
		Tag = TAG_H3;
	} else
	if (strequi(tagtop, _t("h4"), 2)){
		Tag = TAG_H4;
	} else
	if (strequi(tagtop, _t("span"), 4)){
		Tag |= TAG_SPAN;
		tnstr spanclass = getTagAttr(tagtop, end, _t("class"));
		if (spanclass == _t("mw-headline")){
			State = ST_HEADLINE;
		} else
		if (spanclass == _t("editsection")){
			EditSection = true;
		} else
		if (Language.exist()){
			if (spanclass==_t("IPA")){
				State = ST_IPA;
			} else
			if (spanclass==_t("unicode audiolink")){
				// for Wiktionary(EN)
				State = ST_AUDIOLINK;
			}
		}
	} else
	if (strequi(tagtop, _t("button"), 6)){
		// for Wiktionary(FR) audio
		tnstr onclick = getTagAttr(tagtop, end, _t("onclick"));
		if (onclick.exist()){
			const tchar *url = _tcsstr(onclick, _t("http://"));
			if (url){
				const tchar *ext = _tcsstr(url, _t(".ogg"));
				if (ext){
					tnstr audio_url(url, STR_DIFF(ext+4, url));
					setAudioUrl(audio_url);
				}
			}
		}
	}
	if (State==ST_AUDIOLINK){
		if (Language.exist()){
			if (strequi(tagtop, _t("a href"), 6)){
				// audio url
				setAudioUrl(getTagAttr(tagtop, end, _t("href")));
			}
		}
	}
	// tags for contents
	if (Section & SEC_CONTENTS){
		if (strequi(tagtop, _t("ol"), 2)){
			Tag |= TAG_OL;
			LICounter = 0;
		} else
		if (strequi(tagtop, _t("ul"), 2)){
			Tag |= TAG_UL;
		} else
		if (strequi(tagtop, _t("li"), 2)){
			Tag |= TAG_LI;
			if (Tag & TAG_OL){
				tchar buf[20];
				buf[0] = '\n';
				itos(++LICounter, &buf[1]);
				_tcscat(buf, _t("."));
				addText(Text[Language].parts[Part], buf, _tcslen(buf));
			} else
			if (Tag & TAG_UL){
				tchar buf[6];
				_tcscpy(buf, _t("\n - "));
				addText(Text[Language].parts[Part], buf, _tcslen(buf));
			}
		} else
		if (strequi(tagtop, _t("dl"), 2)){
			Tag |= TAG_DL;
		} else
		if (strequi(tagtop, _t("dd"), 2)){
			Tag |= TAG_DD;
			addTextLF(Text[Language].parts[Part]);
		} else
		if (strequi(tagtop, _t("br"), 2)){
			addTextLF(Text[Language].parts[Part]);
		} else
		if (strequi(tagtop, _t("table"), 5)){
			Tag |= TAG_TABLE;
		} else
		if (strequi(tagtop, _t("div"), 3)){
			tnstr divclass = getTagAttr(tagtop, end, _t("class"));
			if (divclass==_t("printfooter")){
				Section = SEC_NONE;	// end of contents
			}
		}
	}
}

void TWiktionaryParser::endTag(const tchar *tagtop, const tchar *end)
{
	if (Tag==TAG_NONE)
		return;

	if (strequi(tagtop, _t("h1"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h2"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h3"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("h4"), 2)){
		Tag = TAG_NONE;
	} else
	if (strequi(tagtop, _t("span"), 4)){
		State = ST_NONE;
		Tag &= ~TAG_SPAN;
	}
	// tags for contents
	if (Section & SEC_CONTENTS){
		if (strequi(tagtop, _t("ol"), 2)){
			Tag &= ~TAG_OL;
		} else
		if (strequi(tagtop, _t("ul"), 2)){
			Tag &= ~TAG_UL;
		} else
		if (strequi(tagtop, _t("li"), 2)){
			Tag &= ~TAG_LI;
		} else
		if (strequi(tagtop, _t("dl"), 2)){
			Tag &= ~TAG_DL;
		} else
		if (strequi(tagtop, _t("dd"), 2)){
			Tag &= ~TAG_DD;
		} else
		if (strequi(tagtop, _t("table"), 5)){
			Tag &= ~TAG_TABLE;
		}
	}
}

void TWiktionaryParser::rawText(const tchar *text, int len)
{
	if (Section==SEC_NONE && Tag==TAG_NONE)
		return;

	tnstr s;
	switch (Tag){
		case TAG_H2|TAG_SPAN:
			if (State==ST_HEADLINE){
				Language.set(text, len);
				Langs.add(Language);
				//printf("lang:%ls\n", Language.c_str());
				if (Language.exist()){
					Section = SEC_LANG;
				} else {
					Section = SEC_NONE;
				}
			}
			break;
		case TAG_H3|TAG_SPAN:
			if (State==ST_HEADLINE){
				//_tprintf(_t("head:%s\n"), tnstr(text,len).c_str());
				if (Language.exist()){
					Part.set(text, len);
					if (Part==_t("Pronunciation")){
						Section = SEC_PRON;
					} else {
						if (Part.exist()){
							Section = SEC_PART|SEC_CONTENTS;
						} else {
							Section = SEC_LANG;
						}
					}
				}
			}
			break;
		case TAG_H4|TAG_SPAN:
			if (State==ST_HEADLINE){
				_tprintf(_t("head:%s\n"), tnstr(text,len).c_str());
				if (Language.exist()){
					s.set(text, len);
					if (s==_t("Synonyms")){
						Section = SEC_SYN|SEC_CONTENTS;
						Part = s;
					} else
					if (s==_t("Antonyms")){
						Section = SEC_ANT|SEC_CONTENTS;
						Part = s;
					} else {
						Section = SEC_LANG;
					}
				}
			}
			break;
		default:
			if (State==ST_IPA){
				if (Language.exist()){
					if (Text[Language].ipa.empty()){
						Text[Language].ipa.set(text, len);	// IPA
						//printf("ipa:%ls\n", Text[Language].ipa.c_str());
					}
				}
			}
			if (Section&SEC_CONTENTS){
				// plain text
				addText(Text[Language].parts[Part], text, len);
			}
			break;
	}
}

void TWiktionaryParser::addText(tnstr &s, const tchar *text, int len)
{
	if (Tag & TAG_TABLE)
		return;
	if (s.exist()){
		tchar last = s[s.length()-1];
		if (text[0]=='\n'){
			if (last=='\n'){
				text++;
				len--;
			}
		} else {
			if (last!=' '){
				//s += _t(" ");
			}
		}
	} else {
		while (text[0]=='\n'){
			text++;
			len--;
		}
	}
	s += UnescapeHtml(text, len);
	//s += tnstr(text, len);
}

void TWiktionaryParser::addTextLF(tnstr &s)
{
	addText(s, _t("\n"), 1);
}

void TWiktionaryParser::setAudioUrl(const tchar *url)
{
	if (Language.empty()) return;
	if (!url[0]) return;
	if (Text[Language].audio_url.empty()){
		Text[Language].audio_url = url;
	}
}

#ifdef _DEBUG
// for debug
void TWiktionaryParser::dumpParsedInfo()
{
	for(pimap_t::iterator it=Text.begin();it!=Text.end();it++){
		_tprintf(_t("----%s\n"), it->first.c_str());
		ParsedInfo &pi = it->second;
		_tprintf(_t("ipa  :%s\n"), pi.ipa.c_str());
		_tprintf(_t("audio:%ls\n"), pi.audio_url.c_str());
		foreach(pi.parts, it, tnstr_map){
			_tprintf(_t("[%ls]\n"), it->first.c_str());
			_tprintf(_t("%ls\n"), it->second.c_str());
		}
	}
}
#endif	// _DEBUG

void TWiktionaryParser::getParsedInfo(tnstrbuf &s, tnstrbuf &pron)
{
	if (!EditSection)
		return;	// no data
	foreach_tnstr_vec(Langs, it){
		ParsedInfo &pi = Text[*it];
		s += _t("[[");
		s += *it;
		s += _t("]]\n");

		if (pi.ipa.exist()){
			s += _t("ipa  :");
			s += pi.ipa;
			//TODO: ŏĂHHs += _t("\n");
		}
		if (FmtAudioUrl){
			if (pi.audio_url.exist()){
				s += _t("\naudio:");
				s += pi.audio_url;
				s += _t("\n");
			}
		}
		foreach(pi.parts, it, tnstr_map){
			if (FmtPartName){
				s += _t("[");
				s += it->first;
				s += _t("]\n");
			}
			s += it->second;
			s += _t("\n");
		}
	}
}

tnstr TWiktionaryParser::getAudioUrl(int index)
{
	if (index<0 || index>=Langs.size())
		return NULL;
	if (Langs[index].empty() || Text[Langs[index]].audio_url.empty())
		return _t("\x01");	// no url
	return Text[Langs[index]].audio_url;
}


int wikRequestData(int index, const tchar *srch_word, TWDResponse *resp, const tchar *dictionary_name, TWDInetAccess *inet)
{
	tnstrA srchU;
	{
		tnstrA temp;
		temp.setUTF8(srch_word);	// UTF-16->UTF-8
		EscapeURIString(temp, srchU);	// escape
	}

	__tnstrA url;
	url += Properties[index].url;
	url += srchU;
	BSTR res;
	if (!inet->HttpReq(url, &res)){
		return 0;
	}
	tnstr text1;
	tnstrbuf text2;
	text1.setUTF8((char*)res);
	SysFreeString(res);

	const tchar *text = text1;
			
	if (*resp->cancelReq){
		return 0;
	}

	bool ok = false;
#if 0
	TOFile file;
	file.create(_t("capture.html"));
	file.putline(text1);
	file.close();
#endif
	TWiktionaryParser parser;
	tnstrbuf text_pron;

	//THtmlStreamParser parser;
	if (parser.openBuffer(text1)){
		parser.parse();
		parser.getParsedInfo(text2, text_pron);
		if (text2.exist()){
			text = text2;
			ok = true;
#ifdef MODULE_TEST
			{
				TOFile file;
				file.create(_t("genout.txt"));
				if (text_pron.exist()){
					file.put(_t("PRON:"));
					file.putline(text_pron);
				}
				file.put(text2);
				file.close();
			}
#endif

			if (AudioDownload){
				// Audio download
				int findex = 0;
				for (int i=0;;i++){
					tnstr audio_url = parser.getAudioUrl(i);
					if (audio_url.empty())
						break;
					if (audio_url[0]==0x01)
						continue;
					// t@C͏㏑
					__tnstrA audio_urlA; audio_urlA.setUTF8(audio_url);
					__tnstrA voiceurl = MakeUrl(url, audio_urlA);
					tnstr voicefile = MakeVoiceFile(dictionary_name, audio_urlA, VOFILE_DIR, VOFILE_PREFIX);
					tnstr linkname = GetFileName(voicefile);

					if (inet->GetHttpFile(voiceurl, "", voicefile)==0
						|| fexist(voicefile)	//Note: voicefileقŎgpŃG[ɂȂ\邽
						){
						resp->filelinks[findex++] = SysAllocString(linkname);
					}
				}
			}
		}
	}

	if (*resp->cancelReq || !ok){
		return 0;
	}

	resp->japa = SysAllocString(text);
	if (text_pron.exist()){
		resp->pron = SysAllocString(text_pron);
	}

	return 1;
}

#define	profInt(keyname, value)	GetPrivateProfileInt(_t("config"), keyname, value, IniName)

void wikInit(HINSTANCE hinst)
{
	const tchar *IniName = _t("WiktionaryWD.ini");
	TCHAR module[300];
	if (GetModuleFileName(hinst, module, sizeof(module)/sizeof(TCHAR))>0){
		GetFileExtension(module)[0] = '\0';
		_tcscat(module, _t("ini"));
		IniName = module;
	}

	AudioDownload = profInt(_t("AudioEnabled"), AudioDownload);
	TWiktionaryParser::FmtPron = profInt(_t("PronEnabled"), TWiktionaryParser::FmtPron);
}

